// BaseObject.cpp: implementation of the BaseObject class.
//
//////////////////////////////////////////////////////////////////////

#include "main.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BaseObject::BaseObject()
{
	// isBase prevents functions of this instance being called
	isBase = false; 
	// isFinal prevents instances or sub-classes to be created
	isFinal = false;
	isCreated = true;

	isPersistant = false; // Only affects nameless classes.
	data = 0;
	base = 0;
	script = 0;
}

BaseObject::~BaseObject()
{
	if(data) delete data;
}


Value BaseObject::callFunction(String &funcName, vector<Value> &params)
{
	if(isBase) throw LangError("Cannot call function of base class", "BaseObject::callFunction()", true);
	return callFunction(funcName, params, this);
}

void BaseObject::construct(BaseObject *obj)
{
	if(base) base->construct(obj);
	return; // By default, no memory is allocated
}

bool BaseObject::isSubClass(BaseObject *obj)
{
	if(!base) return false;
	if(base==obj) return true;

	return base->isSubClass(obj);
}

void BaseObject::cleanupData(BaseObject *context)
{
	return;
}

void BaseObject::cleanupData()
{
	cleanupData(this);
}
