// BaseObject.h: interface for the BaseObject class.
//
// BaseObject is the interface through which all objects are accessed.
// Since both native and scripted classes are derived from this class,
// the script object does not have to distinguish which type of class
// is being referred to.
// 
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_BASEOBJECT_H__9D797CAF_2CCF_4172_94A9_ACF28D93B9D0__INCLUDED_)
#define AFX_BASEOBJECT_H__9D797CAF_2CCF_4172_94A9_ACF28D93B9D0__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#include "reflist.h"

#define CHECKCOUNT(good) if (!good) throw LangError("Bad parameter count", __CURPOS__);

class BaseObject  
{
public:

///////////// MEMBER FUNCTIONS ////////////////

	// Creation / Destruction of instance functions
	void cleanupData();
	virtual void cleanupData(BaseObject *context);
	virtual void construct(BaseObject *obj);

	// Info functions
	virtual bool isSubClass(BaseObject* obj);

	// Scripted function calling functions
	virtual Value callFunction(String &funcName, vector<Value> &params, BaseObject *context) = 0;
	virtual Value callFunction(String &funcName, vector<Value> &params);

	// Constructors / Destructors
	BaseObject();
	virtual ~BaseObject();

///////////// MEMBER VARIABLES ////////////////
	
	// Object info vars
	bool isBase;
	bool isPersistant; // If is persistant, ignored by the garbage collector.
	bool isFinal;
	String name;
	BaseObject *base;

	// State vars
	RefList refList;
	bool isCreated;
	VarList vars;
	void *data;
	
	// Owner var
	Script * script;


///////////// INLINE FUNCTIONS ////////////////

	// callFunction helper macros:
	Value callFunction(const String &funcName)
	{
		String funcNameT=funcName;
		vector<Value> param;
		return callFunction(funcNameT,param);
	}
	Value callFunction(const String &funcName, const Value &val1)
	{
		String funcNameT=funcName;
		vector<Value> param;
		param.push_back(val1);
		return callFunction(funcNameT, param);
	}
	Value callFunction(const String &funcName, const Value &val1, const Value &val2)
	{
		String funcNameT=funcName;
		vector<Value> param;
		param.push_back(val1);
		param.push_back(val2);
		return callFunction(funcNameT, param);
	}
	Value callFunction(const String &funcName, const Value &val1, const Value &val2, const Value &val3)
	{
		String funcNameT=funcName;
		vector<Value> param;
		param.push_back(val1);
		param.push_back(val2);
		param.push_back(val3);
		return callFunction(funcNameT, param);	
	}


	void checkCount(vector<Value> &list, int size)
	{
		
	}


};

#endif // !defined(AFX_BASEOBJECT_H__9D797CAF_2CCF_4172_94A9_ACF28D93B9D0__INCLUDED_)
