// Expression.h: interface for the Expression class.
//
//
// Expression represents an expression tree created from a statement.
//
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EXPRESSION_H__759D1566_B170_402B_87C5_A70541859E31__INCLUDED_)
#define AFX_EXPRESSION_H__759D1566_B170_402B_87C5_A70541859E31__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define EXP_NORMAL		1	// A "normal" expression, IE, not containing a control structure
#define EXP_BLOCK		2	// A block of statements contained in braces
#define EXP_FOR			3
#define EXP_IF			4
#define EXP_RETURN		5
#define EXP_BREAK		6
#define EXP_WHILE		7
#define EXP_SWITCH		8
#define EXP_CONTINUE	9	// continue
#define EXP_CASE		10

#define EXP_ARRAY		11 // This is used for array declarations





class Expression  
{
	int type;
	ExpressionNode * expression;
	vector<Expression *> block;	// Used for EXP_BLOCK
	Expression * subExp; // Used for 'while', 'if', 'for'

	Expression * elseExp; // Used for 'if .. else'

	ExpressionNode * cond;		// Used for 'while' and 'for'
	ExpressionNode * init;		// Used for 'for'
	ExpressionNode * increment;	// Used for 'for'

	
	vector<int> dimensions; // Used for EXP_ARRAY
public:
	void create(Value &val, int dim);
	
	Expression(String exp, BaseObject * obj, vector<String> &localNames);
	virtual ~Expression();

	Value evaluate(BaseObject *obj, vector<Value> &locals);
};

#endif // !defined(AFX_EXPRESSION_H__759D1566_B170_402B_87C5_A70541859E31__INCLUDED_)
