// ExpressionNode.h: interface for the ExpressionNode class.
//
// This reprents an actual expression that can be evaluated. It is 
// in the form of a binary expression tree.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EXPRESSIONNODE_H__7997A8BF_F45D_4AB3_B6E1_3100AC10EF68__INCLUDED_)
#define AFX_EXPRESSIONNODE_H__7997A8BF_F45D_4AB3_B6E1_3100AC10EF68__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


// expArrayDim notes
// This node type must be its own type

enum ExpType
{
	expOp,			// Internal node
	expConst,       // Leaf node
	expLocal,       // Leaf node
	expMember,      // Leaf node
	expSymbol,      // Leaf node
	expParamList    // This is used both for dynamic array dimensions and function parameters

};

class ExpressionNode  
{
	ExpType type;
	//String str;
	union {
		struct {
			ExpressionNode * left;
			ExpressionNode * right; // This holds the expression for a unary operator
		};
		vector<ExpressionNode *> *pList; // Used for expParamList
	};
		

	// Only one of these values will be used at a time
	union {
		Value *val;
		char oper;
		int localIndex;
		int memberIndex;
		String *sym;
	};

	

	void create(String &str, BaseObject *obj, vector<String> &localNames);
	void create(vector<ExpTerm> &list, BaseObject *obj, vector<String> &localNames);
public:

	
	Value evaluate(BaseObject *obj, vector<Value> &locals);
	Value& getLValue(BaseObject *obj, vector<Value> &locals);
	BaseObject *getObject(BaseObject *obj, vector<Value> &locals);

	void makeOper(char oper, int split, vector<ExpTerm> &list, BaseObject *obj, vector<String> &localNames);

/////////////// Ctor/Dtor ///////////////

	ExpressionNode(String &exp, BaseObject *obj, vector<String> &localNames);
	ExpressionNode(vector<ExpTerm> &list, BaseObject *obj, vector<String> &localNames);
	virtual ~ExpressionNode();

/////////////// Member functions ///////////////

	void createTermList(String &exp, vector<ExpTerm> &list, vector<String> &localNames);
	
};


#endif // !defined(AFX_EXPRESSIONNODE_H__7997A8BF_F45D_4AB3_B6E1_3100AC10EF68__INCLUDED_)
