// Object.cpp: implementation of the Object class.
//
//////////////////////////////////////////////////////////////////////
#define START_MEM()  _CrtMemState state1,state2,state3; _CrtMemCheckpoint(&state1); { do { } while(0)
#define END_MEM()  } _CrtMemCheckpoint(&state2); _CrtMemDifference(&state3,&state1,&state2); _CrtMemDumpStatistics(&state3); do { } while(0)

#include "main.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Object::Object(Script *scr) : funcList(0)
{
	script = scr;
	isCreated = false;
}

Object::~Object()
{
	long funcCount = (long)funcList.size();
	for(long i=0; i<funcCount; i++)
	{
		delete funcList[i];
	}
}

Value Object::callFunction(String &name, vector<Value> &params, BaseObject *context)
{
	for(unsigned int i=0; i<funcList.size(); i++)
	{
		if(funcList[i]->name == name)
			return funcList[i]->execute(context, params);
	}

	if(base)
		return base->callFunction(name, params, context);
	BaseObject *sys;
	if(sys = script->getSystem()) 
		return sys->callFunction(name, params, context);
	throw LangError("Function not found", __CURPOS__);
	return Value();
}

// Creates an object from the string representation of it's body
// The line number is used to keep track of what line each statement is in, to facilitate debugging output
void Object::create(String &body, int line)
{

	int pos = 0;

	// We have three parallel arrays, 

	vector<String> functionNames; // Holds a list of function names
	vector<String> functionParam; // Holds a list of function parameter lists
	vector<String> functionBody; // Holds a list of function bodies
	vector<int> functionLineNum; // Holds a list of the starting line of functions
	
	if(base) 
	{
		if(!base->isCreated) throw LangError("Base class must be defined before subclasses", __FILE__ + __LINE__);
		base->vars.getList(vars);
	}
	
//////// Parse declarations ////////

	String word, word2, word3;

	while(pos < body.getLen())
	{
		
		word = body.getWordAt(pos);
		
		// No more words are left, break out
		if(word == "") break;

		if(word == "var")
		{
			pos = body.getLastPos() + 3;
			word = body.getWordAt(pos);
			pos = body.getLastPos() + word.getLen();

			vars.addVar(word); // Add the new variable name
			
			char c = body.getNextChar(pos);

			if(c != ';') 
			{
				String errText = String("Missing ';' after variable declaration in class ") + this->name + ", line " + String::toString(body.getLine(body.getLastPos()) + line);
				throw LangError(errText, __FILE__);
			}

			pos = body.getLastPos() + 1;

		} else if (word == "function")
		{
			

			pos = body.getLastPos() + 8;
			word = body.getWordAt(pos); // Function name
			pos = body.getLastPos() + word.getLen();

			// Error checking, make sure that there is a '(' following the function name
			char c = body.getNextChar(pos);
			if( c != '(' ) 
			{
				String errText = String("Missing '(' at the start of function ") + this->name + "::" + word + "(), line " + String::toString(body.getLine(body.getLastPos()) + line);
				throw LangError(errText, __FILE__);
			}

			// Get function prototype
			try {
				word2 = body.getParenBlockAt(pos); // Function prototype
			} catch (LangError e)
			{
				e.errText = String("Syntax error in function ") + this->name + "::" + word + ", bad paramters";
				throw e;
			}
			// Move pos past prototype
			pos = body.getLastPos() + word2.getLen()+1;

			// Error checking: Make sure that there is a '{' following the function name and parameter list
			c = body.getNextChar(pos);
			if( c != '{' ) 
			{
				String errText = String("Missing '{' at the start of function body ") + this->name + "::" + word + "(), line " + String::toString(body.getLine(body.getLastPos()) + line);
				throw LangError(errText, __FILE__);
			}

			int functionLine = body.getLine(body.getLastPos());

			try {
				word3 = body.getCodeBlockAt(pos); // Function body
			} catch (LangError e)
			{
				e.errText = String("Syntax error in function ") + this->name + "::" + word + ", missing '}' or bad function declaration";
				throw e;
			}
			pos = body.getLastPos() + word3.getLen()+1;
			
			

			functionNames.push_back(word);
			functionParam.push_back(word2);
			functionBody.push_back(word3);
			functionLineNum.push_back(functionLine);
			
		} else {
			String errText = String("Syntax error in class ") + this->name + ", found '" + word + "' while looking for 'function' or 'var' in line " + String::toString(body.getLine(body.getLastPos()) + line);
			throw LangError (errText, __FILE__);

		}
	}

	// Create var list
	//if(varNames) vars = VarList(varNames);
	
	// Create function list
	long funcCount = (long)functionNames.size();
	funcList.resize(funcCount, 0);

	for(long i=0; i < funcCount; i++)
	{
		funcList[i] = new Function(functionNames[i], functionParam[i], functionBody[i], this, functionLineNum[i]);
	}


	//Function *f = funcList[0];
	if(base)
	{
		base->construct(this);
	}
	isCreated = true;
}

// Creates an instance of newBase
void Object::create(BaseObject *newBase)
{
	base = newBase;
	base->vars.getList(vars);
	base->construct(this);
	isCreated = true;
}
