// String.cpp: implementation of the String class.
//
//////////////////////////////////////////////////////////////////////

#include "main.h"
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <ctype.h>
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

String::String()
{
	str=new char[1];
	str[0]=0;
	len=0;
	lastPos=0;
}

String::String(const char *cpy)
{
	if(cpy==NULL) throw LangError("Null Pointer","String::String",true);
	len = (long)strlen(cpy);
	str = new char[len+1];
	strcpy(str, cpy);
	lastPos=0;
}

String::String(const String &cpy)
{
	if(cpy.str==NULL) throw LangError("Uninitialized","String::String",true);
	len = cpy.len;
	str = new char[len+1];
	strcpy(str, cpy.str);
	lastPos=cpy.lastPos;
}

String::String(long bufferSize)
{
	len = bufferSize;
	str = new char[len+1];
}

String::~String()
{
	if(str) delete [] str;
}


void String::allocate(long size)
{
	delete [] str;
	str = new char[size+1];
	len=size;
}

void String::assign(const char *cpy)
{
	if(cpy==NULL) throw LangError("Null Pointer","String::assign",true);
	if(str) delete[] str;
	len = (long)strlen(cpy);
	str = new char[len+1];
	strcpy(str, cpy);	
}

void String::operator =(const String &cpy)
{
	allocate(cpy.len);
	strcpy(str,cpy.str);
}


void String::operator +=(const String &add)
{
	char *temp;
	temp=new char[len+1];
	strcpy(temp,str);
	allocate(add.len+len);
	strcpy(str,temp);
	strcat(str,add.str);
	delete []temp;
}

// This function is slow!
String String::operator +(const String &add)
{
	String ret;
	
	char *temp;
	temp=new char[len+add.len+1];
	strcpy(temp,str);
	strcat(temp,add.str);
	ret.assign(temp);
	
	delete []temp;
	return ret;
}

// This function is REALLY slow!
String String::operator +(const int add)
{
	String ret;

	char *temp;
	char buffer[50];
	sprintf(buffer, "%d", add);

	temp = new char[len + 50];

	strcpy(temp, str);
	strcat(temp, buffer);
	ret.assign(temp);
	delete []temp;
	return ret;
}


int String::operator ==(const String &cmp)
{
	return(strcmp(str,cmp.str)==0);
}

int String::operator ==(const char *cmp)
{
	return(strcmp(str,cmp)==0);
}

int String::operator !=(const String &cmp)
{
	return(strcmp(str,cmp.str)!=0);
}

int String::operator !=(const char *cmp)
{
	return(strcmp(str,cmp)!=0);
}

int String::operator >(const String &cmp)
{
	return(strcmp(str,cmp.str)>0);
}

int String::operator <(const String &cmp)
{
	return(strcmp(str,cmp.str)<0);
}

int String::operator >=(const String &cmp)
{
	return(strcmp(str,cmp.str)>=0);
}

int String::operator <=(const String &cmp)
{
	return(strcmp(str,cmp.str)<=0);
}



String::operator char *()
{
	return str;
}

String::operator const char *() const
{
	return str;
}


void String::split(char splt, StringList &list)
{
	long lastEnd=0;
	long curPos;
	long curStr=0;
	long cnt = countChar(splt);
 	list.resize(cnt+1);
	for(curPos=0; curPos<len+1; curPos++)
	{
		if(str[curPos]==splt || str[curPos]==null)
		{
			String *s=&list[curStr];
			s->allocate(curPos-lastEnd);
			memcpy(s->str,&str[lastEnd],curPos-lastEnd);
			//for(long i=lastEnd; i<curPos; i++)
			//{
				//s->str[i-lastEnd]=str[i];	
			//}
			s->str[curPos-lastEnd]=0;
			lastEnd=curPos+1;
			curStr++;
		}
	}

}

long String::countChar(char target)
{
	long occur=0;
	for(long i=0; (i<len) && (str[i]!=null); i++)
		if(str[i]==target) occur++;
	return occur;
}

void String::loadFile(const char *filename)
{
	
	FILE *fp=fopen(filename,"r");
	
	int size;
	if(!fp) throw LangError("File not found", "String::loadFile()",true);
	fseek(fp,0,SEEK_END);
	size=ftell(fp);
	fseek(fp,0,SEEK_SET);
	char *buf=new char[size+1];
	size = (long)fread(buf,1,size,fp);
	buf[size]=0;
	*this = buf;
	delete [] buf;
	fclose(fp);
}

// trimWhitespace removes multiple spaces, replacing them with a single space. It 
// does not do anything to tabs.
void String::trimWhitespace()
{
	int start, end, i;
	char *temp = new char[len+1];
	
	for(i=0; (i<len) && (isspace(str[i])); i++);
	start = i;
	
	for(i=len-1; (i>start) && (isspace(str[i])); i--);
	end = i;
	for(i=start; i<=end; i++)
		temp[i-start]=str[i];
	temp[end-start+1]=0;

	assign(temp);
	delete [] temp;
}

// Function: getWordAt
// Precondition: pos contains the index at which to start looking
// Postcondition: 
// String has been returned containing the word found. 
// LastPos has been updated to point at the beginning of the word found.
String String::getWordAt(int pos)
{
	if(pos<0 || pos>=len) 
	{
		throw LangError("Index out of bounds", "String::getWordAt(int pos)", true);
	}

	while((pos<len) && (isspace(str[pos])))
	{
		pos++;
	}

	// pos now contains the start of the word
	
	int i=pos;
	lastPos=pos; // LastPos is updated to the start of the word
	while((i<len) && (iscsym(str[i]))) // Increment as long as i is within bounds and current character is a valid char
	{
		i++;
	}

	// i now contains the index to the end of the word

	String ret; // String to be returned
	ret.allocate(i-pos); // resize string to (i - pos) = size of word
	for(int j=pos; j<i; j++)
		ret.str[j-pos]=str[j]; // Copy into the destination string
	ret.str[i-pos]=0; // Set the null pointer at the end of the string
	return ret;
}

String String::getLiteralAt(int pos, int &escCount)
{
	escCount = 0;
	while(isspace(str[pos])) pos++;
	if(str[pos]!='\"') throw LangError("Expecting Literal", "String::getLiteralAt", true);
	pos++;
	int end=pos;
	int count=0;
	while(str[end]!=0)
	{
		if(str[end]=='\\')
		{
			end++;
		}
		if(str[end]=='\"' && str[end-1]!='\\')
			break;
		end++;
		count++;
	}
	
	if(str[end]==0) throw LangError("Unterminated Literal", "String::getLiteralAt", true);
	lastPos=end+1; //lastPos is the end of string
	char *newStr = new char[count+1];
	int cur=0;
	for(; pos<end; pos++)
	{
		if(str[pos]=='\\')
		{
			pos++;
			escCount++;
			if(str[pos]=='\"')
				newStr[cur]='\"';
			if(str[pos]=='\\')
				newStr[cur]='\\';
			if(str[pos]=='n')
				newStr[cur] = '\n';
			if(str[pos]=='t') 
				newStr[cur] = '\t';
			if(str[pos]=='v')
				newStr[cur] = '\v';
			if(str[pos]=='b')
				newStr[cur] = '\b';
			if(str[pos]=='r')
				newStr[cur] = '\r';
			if(str[pos]=='f')
				newStr[cur] = '\f';
			if(str[pos]=='a')
				newStr[cur] = '\a';
			if(str[pos]=='?')
				newStr[cur]= '\?';
		} else {
			newStr[cur]=str[pos];
		}
		cur++;
	}

	newStr[count]=0;
	String ret;
	ret.allocate(end-pos);
	ret.assign(newStr);
	delete [] newStr;
	return ret;
	
}

String String::getSubString(long start, long end) // Returns string from start to end, inclusive
{
	String ret;
	end++;
	ret.allocate(end-start);
	memcpy(ret.str,&str[start],end-start);
	ret.str[end-start]=0;
	lastPos=start;
	ret.len=(long)strlen(ret.str);
	return ret;
}

// Returns the next char, ignoring whitespace
// If it reaches the end of the string, it returns null
char String::getNextChar(int pos) 
{
	if(len == 0) return 0; // Special case for zero length strings. 

	if(pos<0 || pos>=len) 
	{
		throw LangError("Index out of bounds", "String::getNextChar(int pos)", true);
	}
	while((pos<len) && (isspace(str[pos])))
	{
		pos++;
	}
	lastPos=pos;
	return str[pos];
		
}

long String::getLastPos()
{
	return lastPos;
}

// Returns the next code block contained in braces.
// Updates LastPos.
String String::getCodeBlockAt(long pos)
{
	if(pos<0 || pos>=len) throw LangError("Index out of bounds", "String::getCodeBlockAt(long pos)", true);
	while((pos<len) && (isspace(str[pos])))
	{
		pos++;
	}
	if(str[pos]!='{') throw LangError("Bad code block", "String::getCodeBlockAt(long pos)", true);
	lastPos=pos;
	pos++;
	long start=pos;
	int level=1;
	bool isLit=false;
	while((pos<len) && level!=0)
	{
		if(str[pos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[pos-1]!='\\') 
				isLit=false;
		}
		if(str[pos]=='{' && !isLit) level++;
		if(str[pos]=='}' && !isLit) level--;
		pos++;
	}
	if(level!=0) throw LangError("Unterminated code block: Expecting '}'", "String::getCodeBlockAt(long pos)", true);
	return getSubString(start,pos-2);
}

// Returns the next phrase up until the next semicolon.
// Updates LastPos
String String::getStatementAt(long pos)
{
	if(pos<0 || pos>=len) throw LangError("Index out of bounds", "String::getStatementAt(long pos)", true);
	while((pos<len) && isspace(str[pos]))
	{
		pos++;
	}
	lastPos=pos;
	long end=pos;
	int pDepth = 0;
	int bDepth = 0;
	bool isLit=false;
	while(end<len)
	{
		if(str[end]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[end-1]!='\\') 
				isLit=false;
		}
		if(str[end]=='(' && !isLit) pDepth++;
		if(str[end]==')' && !isLit) pDepth--;
		if(str[end]=='{' && !isLit) bDepth++;
		if(str[end]=='}' && !isLit) bDepth--;
		if( (pDepth==0) && (bDepth==0) && 
			(str[end]==';') && !isLit ) break;

		end++;
	}
	
	//while((end<len) && str[end]!=';')
	//{
		//end++;
	//}
	//if(str[end]!=';') throw LangError("Unterminated Statement: expecting ';'", "String::getStatementAt(long pos)", true);
	return getSubString(pos,end-1);
}

String String::getParenBlockAt(long pos)
{
	if(pos<0 || pos>=len) throw LangError("Index out of bounds", "String::getParenBlockAt(long pos)", true);
	while((pos<len) && (isspace(str[pos])))
	{
		pos++;
	}
	if(str[pos]!='(') throw LangError("Bad code block", "String::getParenBlockAt(long pos)", true);
	lastPos=pos;
	pos++;
	long start=pos;
	int level=1;
	bool isLit=false;
	while((pos<len) && level!=0)
	{
		if(str[pos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[pos-1]!='\\') 
				isLit=false;
		}
		if(str[pos]=='(' && !isLit) level++;
		if(str[pos]==')' && !isLit) level--;
		pos++;
	}
	if(level!=0) throw LangError("Unterminated code block: Expecting ')'", "String::getParenBlockAt(long pos)", true);
	return getSubString(start,pos-2);
}

void String::splitIgnoreLiteral(char split, StringList &list)
{
	long lastEnd=0;
	long curPos;
	long curStr=0;
	long cnt = 0;
	bool isLit=false;

	for(curPos=0; curPos<len+1; curPos++)
	{
		if(str[curPos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[curPos-1]!='\\') 
				isLit=false;
		}
		if(str[curPos]==split && !isLit) cnt++;
	}
	

	isLit=false;
 	list.resize(cnt+1);
	for(curPos=0; curPos<len+1; curPos++)
	{
		if(str[curPos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[curPos-1]!='\\') 
				isLit=false;
		}
		if((!isLit && str[curPos]==split) || str[curPos]==null)
		{
			String *s=&list[curStr];
			s->allocate(curPos-lastEnd);
			memcpy(s->str,&str[lastEnd],curPos-lastEnd);
			s->str[curPos-lastEnd]=0;
			lastEnd=curPos+1;
			curStr++;
		}
	}	
	if (isLit) throw LangError("Unterminated String Constant", "String::splitIgonreLiteral()",true);

}

// Returns the next phrase that consists of numeric
// and decimal places or '-' or '+'. Updates LastPos
String String::getNumericAt(int pos)
{
	if(pos<0 || pos>=len) throw LangError("Index out of bounds", "String::getNumericAt(int pos)", true);
	while((pos<len) && (isspace(str[pos])))
	{
		pos++;
	}
	int i=pos;
	lastPos=pos;
	if((str[i]=='-') || (str[i]=='+')) i++;
	while((i<len) && ((isdigit(str[i])) || (str[i]=='.')))
	{
		i++;
	}
	String ret;
	ret.allocate(i-pos);
	for(int j=pos; j<i; j++)
		ret.str[j-pos]=str[j];
	ret.str[i-pos]=0;
	return ret;		
}

void String::splitIgnoreAll(char split, StringList &list)
{
	long lastEnd=0;
	long curPos;
	long curStr=0;
	long cnt = 0;
	bool isLit = false;
	int pDepth = 0;
	int bDepth = 0;
	for(curPos=0; curPos<len+1; curPos++)
	{
		if(str[curPos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[curPos-1]!='\\') 
				isLit=false;
		}
		if(str[curPos]=='(' && !isLit) pDepth++;
		if(str[curPos]==')' && !isLit) pDepth--;
		if(str[curPos]=='{' && !isLit) bDepth++;
		if(str[curPos]=='}' && !isLit) bDepth--;
		if( (pDepth==0) && (bDepth==0) && 
			(str[curPos]==split) && !isLit ) cnt++;
	}
	pDepth=0;
	bDepth=0;
	isLit=false;
 	list.resize(cnt+1);
	for(curPos=0; curPos<len+1; curPos++)
	{
		if(str[curPos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[curPos-1]!='\\') 
				isLit=false;
		}
		if(str[curPos]=='(' && !isLit) pDepth++;
		if(str[curPos]==')' && !isLit) pDepth--;
		if(str[curPos]=='{' && !isLit) bDepth++;
		if(str[curPos]=='}' && !isLit) bDepth--;
		if( (!isLit && str[curPos]==split && bDepth==0 &&
			pDepth==0) || str[curPos]==null )
		{
			String *s=&list[curStr];
			s->allocate(curPos-lastEnd);
			memcpy(s->str,&str[lastEnd],curPos-lastEnd);
			s->str[curPos-lastEnd]=0;
			lastEnd=curPos+1;
			curStr++;
		}
	}	
	if (isLit) throw LangError("Unterminated String Constant", "String::splitIgonreLiteral()",true);	
}

String String::getBracketBlockAt(int pos)
{
	if(pos<0 || pos>=len) throw LangError("Index out of bounds", "String::getParenBlockAt(long pos)", true);
	while((pos<len) && (isspace(str[pos])))
	{
		pos++;
	}
	if(str[pos]!='[') throw LangError("Bad code block", "String::getParenBlockAt(long pos)", true);
	lastPos=pos;
	pos++;
	long start=pos;
	int level=1;
	bool isLit=false;
	while((pos<len) && level!=0)
	{
		if(str[pos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[pos-1]!='\\') 
				isLit=false;
		}
		if(str[pos]=='[' && !isLit) level++;
		if(str[pos]==']' && !isLit) level--;
		pos++;
	}
	if(level!=0) throw LangError("Unterminated code block: Expecting ')'", "String::getParenBlockAt(long pos)", true);
	return getSubString(start,pos-2);
}

String String::getVarDecAt(int pos)
{
	if(pos<0 || pos>=len) 
	{
		throw LangError("Index out of bounds", "String::getWordAt(int pos)", true);
	}
	while((pos<len) && (isspace(str[pos])))
	{
		pos++;
	}
	int i=pos;
	lastPos=pos;
	while((i<len) && (iscsym(str[i])))
	{
		i++;
	}
	i--;
	while(i+1<len && getNextChar(i+1)=='[')
	{
		i++;
		while(str[i]!=']')
		{
			if(str[i]==0)
			{
				throw LangError("Index out of bounds", "String::getVarDecAt()", true);
			}
			i++;
		}
	}
	i++;
	String ret;
	ret.allocate(i-pos);
	for(int j=pos; j<i; j++)
		ret.str[j-pos]=str[j];
	ret.str[i-pos]=0;
	return ret;

}

// Doesn't modify lastPos
void String::splitParamList(vector<String> &list)
{
	int pos = 0;
	long start=0;
	int level=0;
	int plevel=0;
	bool isLit=false;

	while((pos<len))
	{
		if(str[pos]=='\"')
		{
			// If we arent in a literal, we don't have to worry 
			// about an escaped character
			if(!isLit) 
				isLit=true;
			// If we are in a literal, and the previous char is
			// a slash, then the quote is escaped.
			else if(str[pos-1]!='\\') 
				isLit=false;
		}
		if(str[pos]=='[' && !isLit) level++;
		if(str[pos]==']' && !isLit) level--;
		if(str[pos]=='(' && !isLit) plevel++;
		if(str[pos]==')' && !isLit) plevel--;

		if(str[pos]==',' && !isLit && plevel==0 && level==0)
		{
			list.push_back(getSubString(start, pos-1));
			start = pos + 1;
		}

		pos++;
	}

	list.push_back(getSubString(start,getLen()));

	//if(level!=0) throw LangError("Unterminated code block: Expecting ')'", "String::getParenBlockAt(long pos)", true);
	//return getSubString(start,pos-2);

}

 // Returns the number of newline characters from the start of the string to index pos
long String::getLine(int pos)
{
	int count = 0; // Count to be returned
	// Loop from the start of the string to index pos
	for(int i = 0 ; i <= pos; i++)
	{
		if(str[i] == '\n') count++; 
	}

	return count; // Return the number of newline characters found
}

// Gets a block of text specifying array dimensions
// Examples:
// [543][54][4]
// [myVar][2]
// [foobar(3)][otherbar("param")]

String String::getArrayDimensionBlockAt(long pos)
{
	char c = getNextChar(pos);
	if(c != '[') throw LangError("Bad Array Dimension Block", "String::getArrayDimensionBlockAt()", true);
	int start = getLastPos();
	while(1)
	{
		if(c != '[') break;
		pos = getLastPos();

		pos += getBracketBlockAt(pos).getLen() + 2;
        if(pos >= len) break;
		c = getNextChar(pos);

	}

	lastPos = start; // update lastPos variable

	return getSubString(start, pos - 1);
}