// Value1.cpp: implementation of the Value class.
//
//////////////////////////////////////////////////////////////////////

#include "main.h"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Value::Value()
{
	type=NONE;
	s=0;
	flags=0;
}

Value::Value(const Value &cpy)
{
	type = 0;
	assign(cpy);
}

Value::Value(int i_p)
{
	type=INTEGER;
	i = i_p;
	flags=0;
}

Value::Value(float f_p)
{
	type=FLOAT;
	f = f_p;
	flags=0;
}

Value::Value(const String &s_p)
{
	type=STRING;
	s = new String(s_p);
	flags=0;
}

Value::Value(BaseObject *obj_p)
{
	type=OBJECT;
	obj=obj_p;
	obj_p->refList.addRef(this);
	flags=0;
}


Value::~Value()
{
	clear();
}

void Value::assign(String &str)
{
	clear();
	type=NONE;
	if(str.getNextChar(0)=='\"')
	{
		int none;
		s=new String(str.getLiteralAt(0,none));
		type=STRING;
	} else if(str.countChar('.')==1)
	{
		char *ptr;
		String word;
		f = (float)strtod(str, &ptr);
		type=FLOAT;
		word=String(ptr);
		if(word.getLen()!=0 && word.getNextChar(0)!=0) throw LangError("Expecting end of line", "Value::assign(String &str)", true);
		
	} else {
		char *ptr;
		String word;
		i = strtol(str, &ptr, 10);
		type=INTEGER;
		word=String(ptr);
		if(word.getLen()!=0 && word.getNextChar(0)!=0) throw LangError("Expecting end of line", "Value::assign(String &str)", true);
	}

}


void Value::operator =(const Value &cpy)
{
	assign(cpy);
}

void Value::operator =(float newVal)
{
	clear();
	type=FLOAT;
	f=newVal;
}

Value* Value::getElement(int index)
{
	if(type!=ARRAY) throw LangError("Variable type not array!", "Value::operator[]", true);
	if(index<0 || index>=array.count) throw LangError("Index out of bounds", "Value::getElementAt()", true);
	return &(array.list[index]);
}

// Deletes a string (or array) if present
void Value::clear()
{
	
	if(type==STRING)
	{
		delete s;
	} else if(type==ARRAY) {
		delete [] array.list;
	} else if(type==OBJECT) {
		if(obj) obj->refList.removeRef(this);
	}
	
	type=NONE;
	i=0;
}

void Value::operator =(int newVal)
{
	clear();
	type=INTEGER;
	i=newVal;
}

void Value::operator =(const String &newVal)
{
	clear();
	type=STRING;
	s=new String(newVal);
}


void Value::operator =(BaseObject *obj_p)
{
	clear();
	type=OBJECT;
	obj=obj_p;
	obj_p->refList.addRef(this);
}

void Value::assign(const Value *cpy)
{
	assign(*cpy);
}

void Value::assign(const Value &cpy)
{
	clear();
	type=cpy.type;
	if(type==FLOAT)
	{
		f=cpy.f;
	} else if(type==INTEGER)
	{
		i=cpy.i;
	} else if(type==OBJECT)
	{
		obj=cpy.obj;
		obj->refList.addRef(this);
	} else if(type==STRING)
	{
		s=new String(*cpy.s);
	} else if(type==ARRAY)
	{
		type=NONE;
		createArray(cpy.array.count);
		for(int i=0; i<cpy.array.count; i++)
		{
			array.list[i]=cpy.array.list[i];
		}

	}
	flags=cpy.flags;
}

void Value::inc()
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(type==FLOAT) {

		f+=1;

	} else if(type==INTEGER) {

		i++;

	} else if(type==NONE) {

		i=1;
		type=INTEGER;
	} else if(type==STRING) {
		convInt();
		i++;
	} else {
		throw LangError("Attempt to increment not numerical variable", "Value::operator ++()", true);
	}
}

void Value::negate()
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array", __CURPOS__);
	if(type==FLOAT) {

		f= -f;

	} else if(type==INTEGER) {

		i= -i;

	} else if(type==NONE) {
		i=0;
		type=INTEGER;
	} else if(type==STRING) {
		convInt();
		i = -i;
	} else {
		throw LangError("Attempt to negate non-numerical variable", "Value::negate()", true);
	}
}

void Value::dec()
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(type==FLOAT)
	{
		f-=1;
	} else if(type==INTEGER)
	{
		i--;
	} else if(type==STRING) {
		convInt();
		i--;
	} else
	{
		throw LangError("Attempt to decrement not numerical variable", "Value::operator --()", true);
	}
}

void Value::opAdd(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	switch(type)
	{
	case NONE:
		assign(*op);
		break;
	case INTEGER:
		switch(op->type)
		{
		case INTEGER:
			i+=op->i;
			break;
		case FLOAT:
			i+=(long)op->f;
			break;
		case STRING:
			convString();
			*s+=*(op->s);
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case INTEGER:
			f+=op->i;
			break;
		case FLOAT:
			f+=op->f;
			break;
		case STRING:
			convString();
			*s+=*(op->s);
			break;
		}
		break;
	case STRING:
		char buf[250];
		switch(op->type)
		{
		case INTEGER:
			sprintf(buf,"%d",op->i);
			*s+=buf;
			break;
		case FLOAT:
			sprintf(buf,"%f",op->f);
			*s+=buf;
			break;
		case STRING:
			*s+=*(op->s);
			break;
		}
		break;
	}

}




// Converts a FLOAT or INTEGER type value to a STRING
void Value::convString()
{
	char buf[250];
	if(type==INTEGER) {
		sprintf(buf,"%d",i);
	} else if(type==FLOAT) {

		sprintf(buf,"%f",f);
	} else if(type==STRING) {
		return;
	} else if(type==NONE) {
		buf[0]=0;
	} else {
		throw LangError("Bad conversion of type", "Value::convString()", true);
	}
	s=new String(buf);
	type=STRING;
}

bool Value::isTrue()
{
	if((type==INTEGER) && (i!=0)) return true;
	if((type==OBJECT) && (obj!=0)) return true;
	return false;
}

void Value::opEquality(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(type==NONE)
	{
		type=INTEGER;
		i=0;
	}
	
	switch(type)
	{
	case INTEGER:
		switch(op->type)
		{
		case NONE:
			i = (i == 0);
			break;
		case INTEGER:
			i = (i == op->i);
			break;
		case FLOAT:
			i = ((float)i == op->f);
			break;
		case STRING:
			i = 0;
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case NONE:
			i = (f == 0);
			break;
		case INTEGER:
			i = (f == (float)op->i);
			break;
		case FLOAT:
			i = (f == op->f);
			break;
		case STRING:
			i = 0;
			break;
		}
		break;
	case STRING:
		switch(op->type)
		{
		case NONE:
			i = ((*s) == "");
			break;
		case INTEGER:
			i = 0;
			break;
		case FLOAT:
			i = 0;
			break;
		case STRING:
			i = (*s == *(op->s));
			break;
		}
		break;
	}
	type=INTEGER;
}

void Value::opSub(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	switch(type)
	{
	case NONE:
		assign(op);
		break;
	case INTEGER:
		switch(op->type)
		{
		case INTEGER:
			i-=op->i;
			break;
		case FLOAT:
			i-=(long)op->f;
			break;
		case STRING:
			throw LangError("Subtraction of strings not legal", "Value::opSub(Value *op)", true);
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case INTEGER:
			f-=op->i;
			break;
		case FLOAT:
			f-=op->f;
			break;
		case STRING:
			throw LangError("Subtraction of strings not legal", "Value::opSub(Value *op)", true);
			break;
		}
		break;
	case STRING:
		throw LangError("Subtraction of strings not legal", "Value::opSub(Value *op)", true);
		break;
	}
}


void Value::opMult(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	switch(type)
	{
	case NONE:
		if(op->type==STRING) throw LangError("Can't multiply strings", "value::opMult()", true);
		// None is treated like 0 in this case.
		break;
	case INTEGER:
		switch(op->type)
		{
		case INTEGER:
			i*=op->i;
			break;
		case FLOAT:
			i*=(long)op->f;
			break;
		case STRING:
			throw LangError("Can't multiply strings", "value::opMult()", true);
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case INTEGER:
			f*=op->i;
			break;
		case FLOAT:
			f*=op->f;
			break;
		case STRING:
			convString();
			throw LangError("Can't multiply strings", "value::opMult()", true);
			break;
		}
		break;
	case STRING:
		throw LangError("Can't multiply strings", "value::opMult()", true);
		break;
	}
}


void Value::opDiv(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	switch(type)
	{
	case NONE:
		if(op->type==STRING) throw LangError("Can't divide strings", "value::opDiv()", true);
		// None is treated like 0 in this case.
		break;
	case INTEGER:
		switch(op->type)
		{
		case INTEGER:
			if(op->i==0) throw LangError("Division by zero", "value::opDiv()", true);
			convFloat(); // This prevents wierd integer division
			f/=(op->i);
			break;
		case FLOAT:
			if(op->f==0) throw LangError("Division by zero", "value::opDiv()", true);
			i/=(long)op->f;
			break;
		case STRING:
			throw LangError("Can't divide strings", "value::opDiv()", true);
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case INTEGER:
			if(op->i==0) throw LangError("Division by zero", "value::opDiv()", true);
			f/=op->i;
			break;
		case FLOAT:
			if(op->f==0) throw LangError("Division by zero", "value::opDiv()", true);
			f/=op->f;
			break;
		case STRING:
			convString();
			throw LangError("Can't divide strings", "value::opDiv()", true);
			break;
		}
		break;
	case STRING:
		throw LangError("Can't divide strings", "value::opDiv()", true);
		break;
	}
}

void Value::opBitwiseAnd(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(op->type!=INTEGER || type!=INTEGER) throw LangError("Bitwise And can only be used with integers", "value::opBitwiseAnd()",true);
	i &= op->i;
}

void Value::opBitwiseOr(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(op->type!=INTEGER || type!=INTEGER) throw LangError("Bitwise Or can only be used with integers", "value::opBitwiseOr()",true);
	i |= op->i;
}

void Value::opBitwiseXor(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(op->type!=INTEGER || type!=INTEGER) throw LangError("Bitwise Xor can only be used with integers", "value::opBitwiseXor()",true);
	i ^= op->i;
}

void Value::opLogicalAnd(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if( op->isTrue() && isTrue() )
	{
		i=1;
	} else {
		i=0;
	}
	type=INTEGER;
}

void Value::opLogicalOr(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if( op->isTrue() || isTrue() )
	{
		i=1;
	} else {
		i=0;
	}
	type=INTEGER;
}


void Value::opGreater(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(type==NONE)
	{
		type=INTEGER;
		i=0;
	}

	switch(type)
	{
	case STRING:
		if(op->type!=STRING)
		{
			throw LangError("Cannot compare string to non-string", "Value::opGreater()", true);
		}
		if(*s > *(op->s))
		{
			clear();
			i=1;
		} else {
			clear();
			i=0;
		}
		type=INTEGER;
		break;
	case INTEGER:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opGreater()", true);
			break;
		case INTEGER:
			if(i>op->i)
				i=1;
			else
				i=0;
			break;
		case FLOAT:
			if(i>op->f)
				i=1;
			else
				i=0;
			break;
		case NONE:
			if(i>0)
				i=1;
			else
				i=0;
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opGreater()", true);
			break;
		case INTEGER:
			if(f>op->i)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case FLOAT:
			if(f>op->f)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case NONE:
			if(f>0)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		}
		break;
	}
}
void Value::opLess(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(type==NONE)
	{
		type=INTEGER;
		i=0;
	}

	switch(type)
	{
	case STRING:
		if(op->type!=STRING)
		{
			throw LangError("Cannot compare string to non-string", "Value::opLess()", true);
		}
		if(*s < *(op->s))
		{
			clear();
			i=1;
		} else {
			clear();
			i=0;
		}
		type=INTEGER;
		break;
	case INTEGER:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opLess()", true);
			break;
		case INTEGER:
			if(i<op->i)
				i=1;
			else
				i=0;
			break;
		case FLOAT:
			if(i<op->f)
				i=1;
			else
				i=0;
			break;
		case NONE:
			if(i<0)
				i=1;
			else
				i=0;
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opLess()", true);
			break;
		case INTEGER:
			if(f<op->i)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case FLOAT:
			if(f<op->f)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case NONE:
			if(f<0)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		}
		break;
	}

}
void Value::opGreaterOrEqual(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(type==NONE)
	{
		type=INTEGER;
		i=0;
	}

	switch(type)
	{
	case STRING:
		if(op->type!=STRING)
		{
			throw LangError("Cannot compare string to non-string", "Value::opGreaterOrEqual()", true);
		}
		if(*s >= *(op->s))
		{
			clear();
			i=1;
		} else {
			clear();
			i=0;
		}
		type=INTEGER;
		break;
	case INTEGER:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opGreaterOrEqual()", true);
			break;
		case INTEGER:
			if(i>=op->i)
				i=1;
			else
				i=0;
			break;
		case FLOAT:
			if(i>=op->f)
				i=1;
			else
				i=0;
			break;
		case NONE:
			if(i>=0)
				i=1;
			else
				i=0;
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opGreaterOrEqual()", true);
			break;
		case INTEGER:
			if(f>=op->i)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case FLOAT:
			if(f>=op->f)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case NONE:
			if(f>=0)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		}
		break;
	}

}
void Value::opLessOrEqual(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if(type==NONE)
	{
		type=INTEGER;
		i=0;
	}

	switch(type)
	{
	case STRING:
		if(op->type!=STRING)
		{
			throw LangError("Cannot compare string to non-string", "Value::opLessOrEqual()", true);
		}
		if(*s <= *(op->s))
		{
			clear();
			i=1;
		} else {
			clear();
			i=0;
		}
		type=INTEGER;
		break;
	case INTEGER:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opLessOrEqual()", true);
			break;
		case INTEGER:
			if(i<=op->i)
				i=1;
			else
				i=0;
			break;
		case FLOAT:
			if(i<=op->f)
				i=1;
			else
				i=0;
			break;
		case NONE:
			if(i<=0)
				i=1;
			else
				i=0;
			break;
		}
		break;
	case FLOAT:
		switch(op->type)
		{
		case STRING:
			throw LangError("Cannot compare string to non-string", "Value::opLessOrEqual()", true);
			break;
		case INTEGER:
			if(f<=op->i)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case FLOAT:
			if(f<=op->f)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		case NONE:
			if(f<=0)
				i=1;
			else
				i=0;
			type=INTEGER;
			break;
		}
		break;
	}
}

void Value::opShiftRight(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if (op->type!=INTEGER || type!=INTEGER)
	{
		throw LangError("Attempt to shift by non-integer data type", "Value::opShiftRight()", true);
	}
	i >>= op->i;
	
}

void Value::opShiftLeft(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if (op->type!=INTEGER || type!=INTEGER)
	{
		throw LangError("Attempt to shift by non-integer data type", "Value::opShiftLeft()", true);
	}
	i <<= op->i;
}

void Value::opModulo(Value *op)
{
	if(type==OBJECT || type==ARRAY) throw LangError ("Operator attempted on object or array");
	if (op->type!=INTEGER || type!=INTEGER)
	{
		throw LangError("Attempt to modulo non-integer data type", "Value::opModulo()", true);
	}
	i %= op->i;
}

void Value::createArray(String &str)
{
	char buffer[20];
	char cur;
	int i;
	for(i = 1; i<20; i++)
	{
		if(i==20) throw LangError("Bad size for array!", "Value::createArray()", true);

		cur=((const char *)str)[i];		
		if(isdigit(cur))
		{
			buffer[i-1]=cur;
		} else if(cur == ']')
		{
			buffer[i-1]=0;
			break;
		} else
		{
			throw LangError("Expecting numeric constant for array size", "value::createArray()", true);
		}
	}
	
	clear();

	type = ARRAY;
	array.count = strtol(buffer, null, 10);
	array.list = new Value[array.count];
	if(str[i+1] == '[')
	{
		String s = str.getSubString(i+1, str.getLen());
		for(int j=0; j<array.count; j++)
		{
			array.list[j].createArray(s);
		}
	}
	
}

void Value::createArray(int size)
{
	clear();

	type = ARRAY;
	array.count = size;
	array.list = new Value[array.count];
}

// This function is called by a BaseObject when the object
// reaches the end of its lifetime. This is needed to prevent
// a pointer to be dereferenced that is no longer valid.
void Value::objectDestroyed()
{
	if(type != OBJECT || obj==0) throw LangError("Unexpected situation", "Value::objectDestroyed()", true);
	obj=0;
	type=NONE;
}

// Converts the value to an integer
void Value::convInt()
{
	switch(type)
	{
	case INTEGER:
		{
			return;
			break;
		}
	case FLOAT:
		{
			i=(int)f;
			type=INTEGER;
			return;
		}
	case STRING:
		{
			int it=strtol(*s, null, 10);
			clear();
			i=it;
			type=INTEGER;
			return;
		}
	case NONE:
		{
			i=0;
			type=INTEGER;
			return;
		}
	default:
		{
			throw LangError("Bad conversion of type, or type mismatch", "Value::convInt()", true);
		}			
	}
}

// Converts the value to a float
void Value::convFloat()
{
	switch(type)
	{
	case INTEGER:
		{
			f = (float)i;
			type = FLOAT;
			return;
		}
	case FLOAT:
		{
			return;
		}
	case STRING:
		{
			float ft = (float)strtod(*s, null);
			clear();
			f=ft;
			type = FLOAT;
			return;
		}
	case NONE:
		{
			f = 0;
			type = FLOAT;
			return;
		}
	default:
		{
			throw LangError("Bad conversion of type, or type mismatch", "Value::convInt()", true);
		}			
	}
}

void Value::createArray(vector<int> &sizes)
{
	createArray(sizes, 0);
}

void Value::createArray(vector<int> &sizes, int dim)
{
	int mySize = sizes[dim];
	createArray(mySize);
	dim++;
	if(dim < sizes.size())
	{
		for(int i = 0; i < mySize; i++)
		{
			getElement(i)->createArray(sizes, dim);
		}
	}
}