// Value.h: interface for the Value class.
//
// Value encapsulates the idea of a general purpose variable
// in the script that can hold any type of value.
// 
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_VALUE_H__2D8B7261_B171_11D2_A7BD_00502C00951D__INCLUDED_)
#define AFX_VALUE_H__2D8B7261_B171_11D2_A7BD_00502C00951D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class Value  
{
public:
	Value(BaseObject *obj);

////////////////// Ctor/Dtor //////////////////

	Value(const Value &cpy);
	Value(int i);
	Value(float f);
	Value(const String &s);
	Value();
	virtual ~Value();

////////////////// Functions //////////////////
	
	// Conversion functions
	void convInt();
	void convFloat();
	void convString();

	
	Value * getElement(int index);
 

	void objectDestroyed();
	void createArray(int size);
	void createArray(String &str);
	void createArray(vector<int> &sizes);
	void createArray(vector<int> &sizes, int dim);
	

	// Operation functions

	void inc();
	void dec();
	void negate();

	void opModulo(Value *op);
	void opShiftLeft(Value *op);
	void opShiftRight(Value *op);
	
	bool isTrue();
	void opAdd(Value *op);
	void opSub(Value *op);
	void opMult(Value *op);
	void opDiv(Value *op);

	void opEquality(Value *op);
	void opBitwiseAnd(Value *op);
	void opBitwiseOr(Value *op);
	void opBitwiseXor(Value *op);

	void opLogicalAnd(Value *op);
	void opLogicalOr(Value *op);

	void opGreater(Value *op);
	void opLess(Value *op);
	void opGreaterOrEqual(Value *op);
	void opLessOrEqual(Value *op);

	// Assignment functions

	void assign(String &str);
	void assign(const Value &cpy);
	void assign(const Value *cpy);

	// Assignment operators

	void operator =(const Value &cpy);
	void operator =(float newVal);
	void operator =(int newVal);
	void operator =(const String &newVal);
	void operator =(BaseObject *obj);


////////////////// Member Variables //////////////////


	// Type determines what type of data is being stored
	int type;
	// Flags contains special values that allow a codeblock
	// to return or break (signals to the higher level block
	// what has happened)
	int flags;

	// This anonymous union stores the actual data.
	// The valid field is determined by type
	union
	{
		double f;
		long i;
		String *s;
		BaseObject *obj; // An object reference
		struct 	// An array
		{
			Value *list;
			int count;
		} array;
	};

	void clear();

private:
	
};

#endif // !defined(AFX_VALUE1_H__2D8B7261_B171_11D2_A7BD_00502C00951D__INCLUDED_)



